import unittest
import pkmodel as pk
import numpy as np


class ProtocolTest(unittest.TestCase):
    """
    Tests the :class:`Protocol` class.
    """
    def test_create(self):
        """
        Tests Protocol creation.
        """
        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[1, 2],
                             instantaneous=True, dose_times=[0, 1, 2, 3])

        self.assertEqual(dosing.dose_amount, 10)
        self.assertEqual(dosing.subcutaneous, True)
        self.assertEqual(dosing.k_a, 0.3)
        self.assertEqual(dosing.continuous, True)
        self.assertEqual(dosing.continuous_period, [1, 2])
        self.assertEqual(dosing.instantaneous, True)
        self.assertEqual(dosing.dose_times, [0, 1, 2, 3])

    def test_dose_function(self):
        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[1, 2],
                             instantaneous=False)
        dose_0 = dosing.dose_time_function(0)
        dose_1 = dosing.dose_time_function(0.5)
        dose_2 = dosing.dose_time_function(1.1)
        dose_3 = dosing.dose_time_function(1.8)
        dose_4 = dosing.dose_time_function(4)

        self.assertEqual(dose_0, 0)
        self.assertEqual(dose_1, 0)
        self.assertEqual(dose_2, 10)
        self.assertEqual(dose_3, 10)
        self.assertEqual(dose_4, 0)

    def test_both_dose(self):
        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[1, 2],
                             instantaneous=True, instant_doses=[10, 20],
                             dose_times=[0.5, 1.5])

        dose_0 = dosing.dose_time_function(0.5)
        dose_1 = dosing.dose_time_function(1.5)

        self.assertLessEqual(dose_0 - (10 / (0.02 * np.sqrt(2 * np.pi))),
                             0.0001)
        self.assertLessEqual(dose_1 - (10 + 20 / (0.02 * np.sqrt(2 * np.pi))),
                             0.0001)

    def test_change_dose(self):
        dosing = pk.Protocol()

        self.assertEqual(dosing.dose_amount, 1)
        dosing.change_dose(100)
        self.assertEqual(dosing.dose_amount, 100)

    def test_modify_dose_type(self):
        dosing = pk.Protocol()
        self.assertEqual(dosing.subcutaneous, False)
        dosing.modify_dose_type(True, 0.3)
        self.assertEqual(dosing.subcutaneous, True)
        self.assertEqual(dosing.k_a, 0.3)

    def test_make_continuous(self):
        dosing = pk.Protocol()
        self.assertEqual(dosing.continuous, False)
        self.assertEqual(dosing.continuous_period, [0, 0])
        dosing.make_continuous(4, 10)
        self.assertEqual(dosing.continuous, True)
        self.assertEqual(dosing.continuous_period, [4, 10])

    def test_add_dose(self):
        dosing = pk.Protocol(instantaneous=False, instant_doses=[],
                             dose_times=[])
        self.assertEqual(dosing.instantaneous, False)
        self.assertEqual(dosing.dose_times, [])
        self.assertEqual(dosing.instant_doses, [])

        dosing.add_dose(1, 1)
        self.assertEqual(dosing.instantaneous, True)
        self.assertEqual(dosing.dose_times, [1])
        self.assertEqual(dosing.instant_doses, [1])

        dosing.add_dose(5, 1)
        self.assertEqual(dosing.instantaneous, True)
        self.assertEqual(dosing.dose_times, [1, 5])
        self.assertEqual(dosing.instant_doses, [1, 1])



#
# Model class
#

class Model:
    """A Pharmokinetic (PK) model
    Parameters
    ----------

    value: numeric, optional
        an example paramter

    Vc: float
        central compartment volume
    Vps: list of floats
        list of volumes of peripheral compartments
    Qps: list of floats
        list of transition rates between central compartment
        and peripheral compartments

    """
    def __init__(self, Vc, Vps, Qps, CL):
        self.__compartments = []
        self.__central_volume = Vc
        self.__n_compartments = 1  # nb Includes central compartment
        self.__CL = CL
        for Vp, Qp in zip(Vps, Qps):
            self.add_compartment(Vp, Qp)

    def add_compartment(self, Vp=1, Qp=1):
        """
        Add a peripheral compartment to the model.
        """
        self.__n_compartments += 1
        self.__compartments.append({'Vp': Vp, 'Qp': Qp})

    @property
    def Vps(self):
        """
        Volumes of the peripheral compartments.
        """
        return [elem['Vp'] for elem in self.__compartments]

    @property
    def Qps(self):
        """
        Transition rates between central compartment
        and peripheral compartments.
        """
        return [elem['Qp'] for elem in self.__compartments]

    @property
    def Vc(self):
        """
        Volume of the central compartment.
        """
        return self.__central_volume

    @property
    def size(self):
        """
        Returns the number of peripheral compartments.
        """
        return self.__n_compartments

    @property
    def CL(self):
        """
        Returns the clearance/elimination rate from the central compartment.
        """
        return self.__CL



import numpy as np


class Protocol:
    """A Pharmokinetic (PK) protocol

    This class protocol sets up the dosing protocol for a pharmokinetic model.
    This includes setting the amount of dose X ng either continous, multiple,
    or single application. Furthermore sets up the dosing type between
    subcutaneous and intravenous.

    Parameters
    ----------
    dose_amount: numeric, optional, default=1
        This parameter takes in the amount of dose given - X ng

    subcutaneous: logical, optional, default = False
        This parameter is set to False for intavenous dosing and True for
        subcutaneous dosing.

    k_a: numerical, optional, dafault = 1
        This parameter specifies the absorption rate for the subcutaneous
        dosing.

    continuous: logical, optional, default = False
        This parameter specifies whether or not the dose X ng is applied at
        a continuous rate of X ng per hour.

    continuous_period: numerical list, optional, default = [0, 0]
        This parameter specifies the time period over which continuous
        dosing is applied. The first number in the list is the time at which
        continuous dosing begins and the second number is when continuous
        dosing ends.

    instantaneous: logical, optional, default = True
        This parameter specifies whether any instantaneous doses of X ng
        take place.

    dose_times: numerical list, optional, default = [0]
        This parameter is a list of numerics that specify the times at which
        instantaneous doses of X ng are applied.

    instant_doses: numerical list, optional, default = [1].
        This parameter is a list of numerics that specify the doses of X ng
        given instantaneously at the times specified in the dose_times param.

    Methods
    -------
    The method dose_time_function() for a particular time ouputs the dose(t).

    Several other methods have been defined to allow certain modifcations of
    the dosing protocol without the necessity of reinitialising the protocol
    for every modification. Until other functions are supported other
    modifications require reintialising the object of class Protocol.
    Supported methods are listed below:

    change_dose(), modify_dose_type(), make_continuous(), add_dose()

    """
    def __init__(self, dose_amount=1, subcutaneous=False,
                 k_a=1, continuous=False, continuous_period=[0, 0],
                 instantaneous=True, dose_times=[0], instant_doses=[1]):
        self.subcutaneous = subcutaneous
        self.k_a = k_a
        self.dose_amount = dose_amount
        self.continuous = continuous
        self.instantaneous = instantaneous
        self.continuous_period = continuous_period
        self.dose_times = dose_times
        self.instant_doses = instant_doses

    def change_dose(self, dose_amount):
        """

        Paramater: dose_amount: The dosage given - X ng


        This method modfies the dose_amount parameter in the object of protocol
        class it is called on.

        """
        self.dose_amount = dose_amount

    def modify_dose_type(self, subcutaneous, k_a=1):
        """

        Paramater: subcutaneous: boolean, required.
            When set as True this specifies there is subcutaneous dosing, and
            when False it specifies intravenous dosing.
        Paramater: k_a: numeric, optional, default = 1.
            The absorption rate for the subcutaneous dosing.


        This method modfies an object of class Protocol to allow users to
        specifiy whether there is subcutaneous or intravenous dosing.
        Additionally for subcutaneous dosing it allows users to modify the
        value of k_a.

        """
        self.k_a = k_a
        if subcutaneous:
            self.subcutaneous = True
        else:
            self.subcutaneous = False

    def make_continuous(self, time_added, time_removed):
        """

        Paramater: time_added: The time that at which want continuous dosing to
        start.
        Parameter: time_removed: The time at whcih you want continuosu dosing
        to end.


        This method modifies an object of class Protocol to convert the dosing
        protocol to continuous over a user specified time period.

        """
        self.continuous = True
        self.continuous_period[0] = time_added
        self.continuous_period[1] = time_removed

    def add_dose(self, time, dose):
        """

        Paramater: time: numeric, required
            Additional time at which you want there to be an instantaneous
            dose.
        Paramater: dose: numeric, required
            The dose in ng that you want applied at this timepoint.


        This method modifies an object of class Protocol to add an additional
        user specified instantaneous dose of dose ng at time. The method
        also specifies that instantaneous dosing does happen should it not
        already be specified.

        """
        self.dose_times.append(time)
        self.instant_doses.append(dose)
        self.instantaneous = True

    def dose_time_function(self, t):
        """

        Paramater: t: numeric, required.
            The time at which you want dose(t) to be returned.

        Returns: numeric.
            Dose(t) for the specific dosing protocol set up in the object
        of class Protocol.

        """
        dose_t_continuous, dose_t_instant = 0, 0
        dose_width = 0.02

        for dose_size, dose_time in zip(self.instant_doses,
                                        self.dose_times):
            dose_t_instant += easy_gaus(t, dose_time, dose_width) * dose_size \
                * self.instantaneous

        if self.continuous and t < self.continuous_period[1] and \
           t >= self.continuous_period[0]:
            dose_t_continuous = self.dose_amount

        dose_t = dose_t_continuous + dose_t_instant
        return dose_t


def easy_gaus(x, mean, std):
    """
    A function which returns generates a gausssian fucntion from user inputted
    mean and std deviation parameters and returns the value of the function at
    user inputted x.

    Parameter: x: numeric, required
        The value at which you wish the gaussian function to be evaluated.
    Parameter: mean: numeric, required
        The mean or the positon on the x axis at which the gaussian function
        is centred or at its peak.
    Parameter: std: numeric, required
        The standard deviation of the gaussian function.
    Returns: float
        The gaussian function evaluated at x, for the use specified parameters.


    """
    return np.exp(-0.5 * (((x - mean) / std)**2)) / (std * np.sqrt(2 * np.pi))






import unittest
import pkmodel as pk


class ModelTest(unittest.TestCase):
    """
    Tests the :class:`Model` class.
    """
    def test_create(self):
        """
        Tests Model creation.
        """
        model = pk.Model(Vc=2., Vps=[], Qps=[], CL=3.)
        self.assertIsInstance(model, pk.Model)
        self.assertIsInstance(model.Vps, list)
        self.assertIsInstance(model.Qps, list)
        self.assertIsInstance(model.CL, float)
        self.assertEqual(model.size, 1)

    def test_add_compartment(self):
        """
        Tests add_compartment function
        """
        model = pk.Model(Vc=2, Vps=[], Qps=[], CL=3)
        model.add_compartment()
        self.assertEqual(model.size, 2)

    def test_properties(self):
        """
        Tests Vc, Vps, Qps, and CL properties
        """
        model = pk.Model(Vc=2., Vps=[], Qps=[], CL=3.)
        model.add_compartment()
        self.assertEqual(len(model.Vps), model.size - 1)
        self.assertEqual(len(model.Qps), model.size - 1)
        self.assertIsInstance(model.Vc, float)
        self.assertIsInstance(model.CL, float)







"""pkmodel is a Pharmokinetic modelling library.

It contains functionality for creating, solving, and visualising the solution
of Parmokinetic (PK) models

"""
# Import version info
from .version_info import VERSION_INT, VERSION  # noqa

# Import main classes
from .model import Model    # noqa
from .protocol import Protocol    # noqa
from .solution import Solution     # noqa



import unittest
import pkmodel as pk
import matplotlib
import numpy as np


class PipelineTest(unittest.TestCase):
    """
    Tests the :class:`Solution` class.
    Init calls solver and so the output of solver is also tested.
    """

    def test_build_iv(self):
        """
        Tests the whole pipeline can be built;
        - make model
        - make protocol
        - make solution
        for an intravenous dosing protocol.
        """

        model = pk.Model(Vc=2., Vps=[1, 2], Qps=[3, 4], CL=3.)

        dosing = pk.Protocol(dose_amount=10, subcutaneous=False,
                             k_a=0.3, continuous=True,
                             continuous_period=[0.2, 0.6],
                             instantaneous=True, dose_times=[0, .1, .2, .3])

        solution = pk.Solution(model=model, protocol=dosing)
        self.assertEqual(solution.sol.y.shape[0], solution.model.size)
        self.assertEqual(solution.sol.y.shape[1], solution.t_eval.shape[0])
        self.assertIsInstance(solution.t_eval, np.ndarray)
        self.assertIsInstance(solution.y0, np.ndarray)

    def test_build_sc(self):
        """
        Tests the whole pipeline can be built;
        - make model
        - make protocol
        - make solution
        for an subcataneous dosing protocol.
        """

        model = pk.Model(Vc=2., Vps=[1, 2], Qps=[3, 4], CL=3.)

        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[0.2, 0.6],
                             instantaneous=True, dose_times=[0, .1, .2, .3])

        solution = pk.Solution(model=model, protocol=dosing)
        self.assertEqual(solution.sol.y.shape[0], solution.model.size + 1)
        self.assertEqual(solution.sol.y.shape[1], solution.t_eval.shape[0])
        self.assertIsInstance(solution.t_eval, np.ndarray)
        self.assertIsInstance(solution.y0, np.ndarray)

    def test_all_zeros(self):
        """
        Tests the whole pipeline;
        - make model
        - make protocol
        - make solution
        for an intravenous dosing protocol.
        """

        model = pk.Model(Vc=2., Vps=[], Qps=[], CL=3.)

        dosing = pk.Protocol(dose_amount=0, subcutaneous=True,
                             k_a=0.3, continuous=False,
                             continuous_period=[],
                             instantaneous=True, dose_times=[])

        solution = pk.Solution(model=model, protocol=dosing)
        sol = solution.sol

        assert(not((sol.y != 0).any()))

    def test_plot_pipeline(self):
        """
        Tests the whole pipeline;
        - make model
        - make protocol
        - make solution
        - plot solution
        for an intravenous dosing protocol.
        """

        model = pk.Model(Vc=2., Vps=[], Qps=[], CL=3.)

        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[0.2, 0.6],
                             instantaneous=True,
                             dose_times=[0, .1, .2, .3])

        solution = pk.Solution(model=model, protocol=dosing)
        sol_fig = solution.generate_plot()
        self.assertIsInstance(sol_fig, matplotlib.figure.Figure)

    def test_compare_pipeline(self):
        """
        Tests the whole pipeline;
        - make model x2
        - make protocol x2
        - make solution x2
        - plot and compare the solutions
        for an intravenous dosing protocol.
        """

        model = pk.Model(Vc=2., Vps=[3, 1], Qps=[1, 3], CL=3.)

        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[0.2, 0.6],
                             instantaneous=True,
                             dose_times=[0, .1, .2, .3])

        solution = pk.Solution(model=model, protocol=dosing)

        model2 = pk.Model(Vc=2., Vps=[4, 5], Qps=[8, 9], CL=3.)

        dosing2 = pk.Protocol(dose_amount=10, subcutaneous=True,
                              k_a=0.3, continuous=True,
                              continuous_period=[0.2, 0.6],
                              instantaneous=True,
                              dose_times=[0, .1, .2, .3])

        solution2 = pk.Solution(model=model2, protocol=dosing2)

        sol_fig = solution.generate_plot(compare=solution2)
        self.assertIsInstance(sol_fig, matplotlib.figure.Figure)

    def test_compare_pipeline_separate(self):
        """
        Tests the whole pipeline;
        - make model x2
        - make protocol x2
        - make solution x2
        - plot and compare the solutions using the separate
        for an intravenous dosing protocol.
        """

        model = pk.Model(Vc=2., Vps=[3, 1], Qps=[1, 3], CL=3.)

        dosing = pk.Protocol(dose_amount=10, subcutaneous=True,
                             k_a=0.3, continuous=True,
                             continuous_period=[0.2, 0.6],
                             instantaneous=True,
                             dose_times=[0, .1, .2, .3])

        solution = pk.Solution(model=model, protocol=dosing)

        model2 = pk.Model(Vc=2., Vps=[4, 5], Qps=[8, 9], CL=3.)

        dosing2 = pk.Protocol(dose_amount=15, subcutaneous=True,
                              k_a=0.3, continuous=True,
                              continuous_period=[0.3, 0.35],
                              instantaneous=True,
                              dose_times=[0, .1, .2, .3])

        solution2 = pk.Solution(model=model2, protocol=dosing2)

        sol_fig = solution.generate_plot(compare=solution2, separate=True)
        self.assertIsInstance(sol_fig, matplotlib.figure.Figure)



import unittest
from unittest.mock import Mock
import pkmodel as pk
import numpy as np
import matplotlib


class SolutionTest(unittest.TestCase):
    """
    Tests the :class:`Solution` class.
    Init calls solver and so the output of solver is also tested.
    """
    def test_intravenous(self):
        """
        Tests Solution creation for intravenous protocol.
        """
        mock_model = Mock()
        mock_model.size = 2
        mock_model.Vc = 1
        mock_model.CL = 1
        mock_model.Qps = [1]
        mock_model.Vps = [1]
        mock_protocol = Mock()
        mock_protocol.subcutaneous = False
        mock_protocol.dose_time_function.return_value = 1

        solution = pk.Solution(model=mock_model, protocol=mock_protocol)
        self.assertEqual(solution.sol.y.shape[0], solution.model.size)
        self.assertEqual(solution.sol.y.shape[1], solution.t_eval.shape[0])
        self.assertIsInstance(solution.t_eval, np.ndarray)
        self.assertIsInstance(solution.y0, np.ndarray)

    def test_subcutaneous(self):
        """
        Tests Solution creation for subcataneous protocol.
        """
        mock_model = Mock()
        mock_model.size = 2
        mock_model.Vc = 1
        mock_model.CL = 1
        mock_model.Qps = [1]
        mock_model.Vps = [1]
        mock_protocol = Mock()
        mock_protocol.subcutaneous = True
        mock_protocol.k_a = 1
        mock_protocol.dose_time_function.return_value = 1

        solution = pk.Solution(model=mock_model, protocol=mock_protocol)
        self.assertEqual(solution.sol.y.shape[0], solution.model.size + 1)
        self.assertEqual(solution.sol.y.shape[1], solution.t_eval.shape[0])
        self.assertIsInstance(solution.t_eval, np.ndarray)
        self.assertIsInstance(solution.y0, np.ndarray)

    def test_plotting(self):
        """
        Tests Solution for the ability to create separate plots for solutions.
        """
        mock_model = Mock()
        mock_model.size = 2
        mock_model.Vc = 1
        mock_model.CL = 1
        mock_model.Qps = [1]
        mock_model.Vps = [1]
        mock_protocol = Mock()
        mock_protocol.subcutaneous = False
        mock_protocol.dose_time_function.return_value = 1

        solution = pk.Solution(model=mock_model, protocol=mock_protocol)

        sol_fig = solution.generate_plot()
        self.assertIsInstance(sol_fig, matplotlib.figure.Figure)

    def test_plotting_separate(self):
        """
        Tests Solution for the ability to create separate plots for solutions.
        """
        mock_model = Mock()
        mock_model.size = 2
        mock_model.Vc = 1
        mock_model.CL = 1
        mock_model.Qps = [1]
        mock_model.Vps = [1]
        mock_protocol = Mock()
        mock_protocol.subcutaneous = False
        mock_protocol.dose_time_function.return_value = 1

        solution = pk.Solution(model=mock_model, protocol=mock_protocol)

        sol_fig = solution.generate_plot(separate=True)
        self.assertIsInstance(sol_fig, matplotlib.figure.Figure)




#
# Solution class
#
import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate


class Solution:
    """A Pharmokinetic (PK) model solution

    Parameters
    ----------

    value: numeric, optional
        an example paramter

    model: using model class
        with model( Vc, Vps, Qps, CL)

    protocol: using protocoll class to specify
        intravenous or subcutaneous dosing
        dosing function needed

    tmax: float
        integrates until it reaches tmax
        default value is 1

    nsteps: int
        number of integration steps
        default value is 1000

    """
    def __init__(self, model, protocol, tmax=1, nsteps=1000):
        self.model = model
        self.protocol = protocol
        self.t_eval = np.linspace(0, tmax, nsteps)
        self.tmax = tmax
        self.nsteps = nsteps

        self.solver()

    def rhs_intravenous(self, t, y):
        '''
        Right hand side of flux equation for intravenous dosing protocol
        dimension of system = number of compartments = self.model.size
        only central compartment dim = 1
        one periphal comartment dim = 2
        two periphal compartment dim = 3
        Parameters
        ----------
        t: time
        y: state vector [qc, q_p1, q_p2]
        '''
        state = y
        Vc = self.model.Vc  # volume of the main compartment
        CL = self.model.CL  # Clearance rate

        cleared = state[0] / Vc * CL  # flux going out of the main compartment
        dq_dt = [0]  # [dqc, dq_p1, dq_p2]

        flux_sum = 0  # sum of flux between compartments
        # loop over peripheral compartments
        for comp in range(1, self.model.size):
            # transition rate etween main compartment
            # and i-th peripheral compartment
            Q_pi = self.model.Qps[comp - 1]
            # volume of i-th peripheral compartment
            V_pi = self.model.Vps[comp - 1]
            # flux between peripheral and main compartment
            flux = Q_pi * (state[0] / Vc - state[comp] / V_pi)
            dq_dt.append(flux)
            flux_sum += flux
        dq_dt[0] = self.protocol.dose_time_function(t) - cleared - flux_sum
        return dq_dt

    def rhs_subcutaneous(self, t, y):
        '''
        Right hand side of flux equation for subcutaneous dosing protocol
        dimension of system = number of compartments = self.model.size
        Parameters
        ----------
        t: time
        y: state vector [q0, qc, q_p1, q_p2]
        y has one dim more than in intravenous protocol
        q0 is an additional compartment from which the drug
        is absorbed to the central compartment
        '''
        state = y
        Vc = self.model.Vc  # volume of the main compartment
        CL = self.model.CL  # Clearance rate

        cleared = state[0] / Vc * CL  # flux going out of the main compartment
        dq_dt = np.zeros(self.model.size + 1)  # [dqc, dq_p1, dq_p2, dq0]
        dq0_dt = self.protocol.dose_time_function(t)
        dq0_dt -= self.protocol.k_a * state[-1]
        dq_dt[-1] = dq0_dt

        flux_sum = 0  # sum of flux between compartments
        # loop over peripheral compartments
        for comp in range(1, self.model.size):
            # transition rate etween main compartment
            # and i-th peripheral compartment
            Q_pi = self.model.Qps[comp - 1]
            # volume of i-th peripheral compartment
            V_pi = self.model.Vps[comp - 1]
            flux = Q_pi * (state[0] / Vc - state[comp] / V_pi)
            dq_dt[comp] = flux
            flux_sum += flux
        dq_dt[0] = self.protocol.k_a * state[-1] - cleared - flux_sum
        return dq_dt

    def solver(self):
        '''
        Runge-Kutta solver
        Dosing protocol specified in protocol class
        determines which rhs function will be used
        '''
        if self.protocol.subcutaneous:
            step_func = self.rhs_subcutaneous
            # subcutaneous protocol has one more dimension
            # than intravenous protocol
            # initial condition y0
            self.y0 = np.zeros(self.model.size + 1)
            self.sol = np.zeros(self.model.size + 1)
        else:
            step_func = self.rhs_intravenous
            # intial condition
            self.y0 = np.zeros(self.model.size)
            self.sol = np.zeros(self.model.size)

        sol = scipy.integrate.solve_ivp(
            fun=lambda t, y: step_func(t, y),
            t_span=[self.t_eval[0], self.t_eval[-1]],
            y0=self.y0, t_eval=self.t_eval,
            max_step=self.tmax / self.nsteps
        )
        self.sol = sol
        return sol

    def plot(self, separate=False):
        """
        Generate a figure of the drug quantity per
        compartment over time for the corresponding model

        :param separate: set to True if you want 1 plot per compartment
        :returns: matplotlib figure
        """
        sol = self.solver()
        n = self.model.size
        if separate:
            fig = plt.figure(figsize=(n * 4.0, 3.0))
            central = fig.add_subplot(1, n, 1)
            central.plot(sol.t, sol.y[0, :], label='- q_c')
            central.legend()
            central.set_title('Central compartment')
        else:
            fig = plt.figure(figsize=(4.0, 3.0))
            model = fig.add_subplot(1, 1, 1)
            model.plot(sol.t, sol.y[0, :], label='- q_c')

        # add legend and axes labels
        plt.ylabel('drug mass [ng]')
        plt.xlabel('time [h]')

        # loop over peripheral compartments and plot drug quantity for each
        for i in range(n - 1):
            label = '- q_p' + str(i + 2)
            if separate:
                subplot = fig.add_subplot(1, n, i + 2)
                subplot.plot(sol.t, sol.y[i + 1, :], label=label)
                subplot.legend()
                subplot.set_xlabel('time [h]')
                subplot.set_title('Peripheral compartment #' + str(i + 1))
            else:
                model.plot(sol.t, sol.y[i + 1, :], label=label)

        # plot subcutanous injections compartment
        if self.protocol.subcutaneous and not separate:
            model.plot(sol.t, sol.y[-1, :], label='- q_0')

        plt.legend()
        fig.tight_layout()
        return fig

    def compare_plots(self, solution_2):
        """
        Generates a matplotlib figure with two subplots that show
        the drug quantity in each compartment over time for the models
        of self and solution_2

        :param solution_2: Should be a Solution object different from self

        :returns: Matplotlib Figure object
        """
        sol1 = self.solver()
        sol2 = solution_2.solver()
        n = max(self.model.size, solution_2.model.size)
        fig = plt.figure(figsize=(2 * 4.0, 3.0))
        model1 = fig.add_subplot(1, 2, 1)
        model1.set_title('Model 1')
        model1.set_xlabel('time [h]')
        model1.set_ylabel('drug mass [ng]')
        model2 = fig.add_subplot(1, 2, 2)
        model2.set_title('Model 2')
        model2.set_xlabel('time [h]')
        for i in range(n):
            if i == 0:
                label = '- q_c'
            else:
                label = '- q_p' + str(i + 1)
            if i < len(sol1.y):
                model1.plot(sol1.t, sol1.y[i, :], label=label)
            if i < len(sol2.y):
                model2.plot(sol2.t, sol2.y[i, :], label=label)
        # plot subcutanous injections compartment
        if self.protocol.subcutaneous:
            model1.plot(sol1.t, sol1.y[-1, :], label='- q_0')
        if solution_2.protocol.subcutaneous:
            model2.plot(sol2.t, sol2.y[-1, :], label='- q_0')
        model1.legend()
        model2.legend()
        fig.tight_layout()
        return fig

    def compare_separate(self, solution_2):
        """
        Generates a matplotlib figure that shows the drug quantity over time
        for the models self and solution_2, with one plot per compartment

        :param solution_2: Should be a Solution object different from self

        :returns: Matplotlib Figure object
        """
        sol1 = self.solver()
        sol2 = solution_2.solver()
        n = max(self.model.size, solution_2.model.size)
        fig = plt.figure(figsize=(n * 4.0, 3.0))
        central = fig.add_subplot(1, n, 1)
        central.plot(sol1.t, sol1.y[0, :], label='model 1')
        central.plot(sol2.t, sol2.y[0, :], label='model 2')
        central.legend()
        central.set_xlabel('time [h]')
        central.set_ylabel('drug mass [ng]')
        central.set_title('Central compartment')
        for i in range(n - 1):
            compartment = fig.add_subplot(1, n, i + 2)
            if i + 1 < len(sol1.y):
                compartment.plot(sol1.t, sol1.y[i + 1, :], label='model 1')
            if i + 1 < len(sol2.y):
                compartment.plot(sol2.t, sol2.y[i + 1, :], label='model 2')
            compartment.legend()
            compartment.set_xlabel('time [h]')
            compartment.set_title('Peripheral compartment #' + str(i + 1))
        fig.tight_layout()
        return fig

    def generate_plot(self, compare=None,
                      separate=False, show=False, savefig=False):
        """
        Calls appropriate function to generate plots of the drug
        quantity per compartment over time for the corresponding model

        :param compare: If None (default), function will only generate plot
        for the Solution object. If set to a Solution object, function will
        generate plots to compare the two models. Else, function will raise an
        Assertion Error.

        :param separate: If False (default), will show all compartments on the
        same plot. Set to True if you want 1 plot per compartment.
        """
        fig = None
        if compare is None:
            fig = self.plot(separate=separate)
        else:
            assert type(compare) is Solution, 'compare should be a Solution'
            if not separate:
                fig = self.compare_plots(compare)
            elif separate:
                fig = self.compare_separate(compare)

        if show:
            plt.show()

        if type(savefig) is str:
            fig.savefig(savefig + '.pdf')
        elif savefig:
            fig.savefig('pkplot.pdf')

        return fig



#
# pkmodel setuptools script
#
from setuptools import setup, find_packages


def get_version():
    """
    Get version number from the pkmodel module.

    The easiest way would be to just ``import pkmodel ``, but note that this may
    fail if the dependencies have not been installed yet. Instead, we've put
    the version number in a simple version_info module, that we'll import here
    by temporarily adding the oxrse directory to the pythonpath using sys.path.
    """
    import os
    import sys

    sys.path.append(os.path.abspath('pkmodel'))
    from version_info import VERSION as version
    sys.path.pop()

    return version


def get_readme():
    """
    Load README.md text for use as description.
    """
    with open('README.md') as f:
        return f.read()


# Go!
setup(
    # Module name (lowercase)
    name='dtp_pkmodel',

    # Version
    version=get_version(),

    description='An example Python project.',

    long_description=get_readme(),

    long_description_content_type="text/markdown",

    license='MIT license',

    # author='',

    # author_email='',

    maintainer='Martin Robinson',

    maintainer_email='martin.robinson@cs.ox.ac.uk',

    url='https://github.com/SABS-R3/2020-software-engineering-projects-pk',

    # Packages to include
    packages=find_packages(include=('pkmodel', 'pkmodel.*')),

    # List of dependencies
    install_requires=[
        # Dependencies go here!
        'numpy>=1.19.2',
        'matplotlib>=3.3.2',
        'scipy>=1.5.3'
    ],
    extras_require={
        'docs': [
            # Sphinx for doc generation. Version 1.7.3 has a bug:
            'sphinx>=1.5, !=1.7.3',
            # Nice theme for docs
            'sphinx_rtd_theme',
        ],
        'dev': [
            # Flake8 for code style checking
            'flake8>=3',
        ],
    },
)
